(* Euclide.ml
   Denis Monasse 1998.
   monasse@cicrp.jussieu.fr
*)

#open "graphics";;

(*
type point = {xc : float; yc : float};;
type droite = {u : float; v : float; h : float};;
type cercle = {xcentre : float;ycentre : float;rayon : float};;

type objet_geometrique =
     Point of point | Droite of droite | Cercle of cercle
     | Triangle of (point*point*point) | Couleur of int | Epaisseur of int;;
*)

exception Erreur of string;;

let sqr x = x *. x;;

(* point x y
   retourne le point de coordonnees (x,y) *)
let point x y = Point {xc=x; yc=y};;

(* droite u v h
   retourne la droite d'equation ux+vy+h=0 *)
let droite u v h = Droite {u=u; v=v; h=h};;

(* cercle x y r
   retourne le cercle de rayon r dont le centre est le point (x,y) *)
let cercle x y r = Cercle {xcentre=x; ycentre=y; rayon =r};;

let axe_Ox = Droite {u=0.; v=1.; h=0.};;
let axe_Oy = Droite {u=1.; v=0.; h=0.};;

(* random_point x1 x2 y1 y2
   retourne un point au hasard dans la fenetre delimitee
   par les abscisses x1 et x2  et les ordonnees y1 et y2 *)
let random_point debutx finx debuty finy =
 let dx = finx -. debutx
 and dy = finy -. debuty in
 Point {xc = (random__float dx) +. debutx;
        yc = (random__float dy) +. debuty};;

(* droite_par_2_points m1 m2  (ou m1 *? m2)
   retourne la droite passant par les points m1 et m2 *)
let droite_par_2_points_aux m1 m2 =
    {u = m1.yc  -.  m2.yc; v = -. m1.xc  +.  m2.xc;
     h = -.  m1.yc *. m2.xc  +.  m1.xc *. m2.yc};;
let droite_par_2_points ob1 ob2 =
    match (ob1,ob2) with
    | (Point m1,Point m2) -> Droite (droite_par_2_points_aux m1 m2)
    | _ -> raise (Erreur "non defini");;

let distance {xc=a1; yc=b1} {xc=a2; yc=b2} =
    sqrt (sqr (a2 -. a1) +. sqr (b2 -. b1));;

let normalise {u=u; v=v; h=h} =
    let norme = sqrt((sqr u) +. (sqr v)) in
    {u=u /. norme; v=v /. norme; h=h /. norme};;


(* point_objet obj t
   retourne le point de parametre t de la droite ou du cercle obj *)
let point_objet ob t =
    let point_de_droite {u=u; v=v;h=h} t =
        let norme = sqrt(u *. u +. v *. v) in
        let norme2 = norme *. norme in
           { xc = ( -. h *. u -. v *. t *. norme) /. norme2;
             yc = ( -. h *. v +. u *. t *. norme) /. norme2 }
    and point_de_cercle c t =
      {xc = c.xcentre +. c.rayon *. cos(t); yc = c.ycentre +. c.rayon *. sin(t)}
    in
    match ob with
    | Droite d -> Point (point_de_droite d t)
    | Cercle c -> Point (point_de_cercle c t)
    | _ -> raise (Erreur "point_objet non defini sur ce type d'objet");;

(* centre_de_cercle c
   retourne le point centre du cercle c *)
let centre_de_cercle = function
    | (Cercle c)  -> Point {xc=c.xcentre; yc=c.ycentre}
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* rayon_de_cercle c
   retourne le rayon du cercle c *)
let rayon_de_cercle = function
    | (Cercle c)  -> c.rayon
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;




(* cercle_par_centre_et_rayon m r
   retourne le cercle ayant pour centre le point m et de rayon r *)
let cercle_par_centre_et_rayon_aux omega r =
    {xcentre=omega.xc; ycentre=omega.yc;rayon=r};;
let cercle_par_centre_et_rayon ob r =
    match ob with
    | Point omega ->
        Cercle {xcentre=omega.xc; ycentre=omega.yc;rayon=r}
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;



(* cercle_par_centre_et_point m1 m2
   retourne le cercle ayant pour centre le point m1
   et passant par le point m2 *)
let cercle_par_centre_et_point_aux omega m =
    {xcentre=omega.xc; ycentre=omega.yc;
     rayon = sqrt((sqr (omega.xc -. m.xc)) +. (sqr(omega.yc -. m.yc)))};;

let cercle_par_centre_et_point m1 m2 =
    match (m1,m2) with
    | (Point omega,Point m) -> Cercle (cercle_par_centre_et_point_aux omega m)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;



(* axe_radical c1 c2  (ou c1 +? c2)
   retourne la droite qui est l'axe radical des deux cercles c1 et c2 *)
let axe_radical_aux
    {xcentre=a1;ycentre=b1;rayon=r1}
    {xcentre=a2;ycentre=b2;rayon=r2} =
      {u=a1 -. a2; v=b1 -. b2;
       h=sqr(r2) -. sqr(r1) +. sqr(a1) -. sqr(a2) +. sqr(b1) -. sqr(b2)};;

let axe_radical c1 c2 =
    match (c1,c2) with
    | (Cercle c'1,Cercle c'2) -> Droite (axe_radical_aux c'1 c'2)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* point_d_intersection d1 d2  (ou d1 *? d2)
   retourne le point d'intersection des deux droites d1 et d2
   points_d_intersection obj1 obj2  (ou  obj1 **? obj2)
   retourne un vecteur dont les deux composantes sont les points
   d'intersection des objets obj1 et obj2, l'un etant un cercle,
   l'autre etant soit un cercle soit une droite  *)
let intersection_d_d d1 d2 =
    let delta = d1.v *. d2.u -. d1.u *. d2.v in
    { xc = (d1.h *. d2.v -. d1.v *. d2.h) /. delta;
      yc = -. (d1.h *. d2.u -. d1.u *. d2.h) /. delta };;

let intersection_d_c {u=u; v=v;h=h} {xcentre=a;ycentre=b;rayon=r} =
    let delta= (sqr a) *. ((sqr u) +. (sqr v)) -. sqr(u *. a +. v *. b +. h)
    and b1 = -. b *. u *. v +. a *. v *. v -. u *. h
    and b2 = -. a *. u *. v +. b *. u *. u -. v *. h
    and den = u *. u +. v *. v in
    if delta <. 0. then raise (Erreur "pas d'intersection")
    else let rdelta = sqrt(delta) in
       ({xc=(b1 +. v *. rdelta) /. den;yc=(b2 -. u *. rdelta) /. den},
       {xc=(b1 -. v *. rdelta) /. den;yc=(b2 +. u *. rdelta) /. den});;

let intersection_c_c c1 c2 = intersection_d_c (axe_radical_aux c1 c2) c1;;

let point_d_intersection ob1 ob2 =
    match (ob1,ob2) with
    | (Droite d1,Droite d2) -> Point (intersection_d_d d1 d2)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let points_d_intersection ob1 ob2 =
    match (ob1,ob2) with
    | (Cercle c1, Droite d2) ->
          let (m1,m2)=intersection_d_c d2 c1 in [| Point m1; Point m2 |]
    | (Droite d1, Cercle c2) ->
          let (m1,m2)=intersection_d_c d1 c2 in [| Point m1; Point m2 |]
    | (Cercle c1, Cercle c2) ->
          let (m1,m2)=intersection_c_c c1 c2 in [| Point m1; Point m2 |]
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* triangle m1 m2 m3
   retourne le triangle ayant pour sommets les points m1, m2, m3
   triangle d1 d2 d3  retourne le triangle dont les cotes sont portes
   par les droites d1, d2 et d3 *)
let triangle ob1 ob2 ob3 =
    match (ob1,ob2,ob3) with
    | ((Point m1) ,(Point m2) ,(Point m3)) -> Triangle (m1,m2,m3)
    | ((Droite d1), (Droite d2), (Droite d3)) ->
          let m1=intersection_d_d d2 d3
          and m2=intersection_d_d d3 d1
          and m3=intersection_d_d d1 d2
          in Triangle (m1,m2,m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* sommets triang
   retourne un vecteur dont les trois composantes sont les
   points sommets du triangle *)
let sommets  = function
    | (Triangle  (m1,m2,m3)) -> [| Point m1; Point m2; Point m3 |]
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* droite_parallele m d   (ou m //? d)
   retourne la droite parallele a d passant par le point m;
   les deux parametres peuvent etre echanges sans inconvenient *)
let droite_parallele_aux {u=u;v=v;h=h} {xc=a;yc=b} =
 {u=u; v=v; h= -. u *. a -. v *. b};;

let droite_parallele o1 o2 =
    match (o1,o2) with
    | (Droite d, Point m) -> Droite (droite_parallele_aux d m)
    | (Point m, Droite d) -> Droite (droite_parallele_aux d m)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;


(* droite_perpendiculaire m d  (ou m +? d)
   retourne la droite perpendiculaire a d passant par le point m;
   les deux parametres peuvent etre echanges sans inconvenient *)
let droite_perpendiculaire_aux {u=u;v=v;h=h} {xc=a;yc=b} =
 {u= -. v; v=u; h=v *. a -. u *. b};;

let droite_perpendiculaire o1 o2 =
    match (o1,o2) with
    | (Droite d, Point m) -> Droite (droite_perpendiculaire_aux d m)
    | (Point m, Droite d) -> Droite (droite_perpendiculaire_aux d m)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* milieu m1 m2
   retourne le point milieu des deux points m1 et m2 *)
let milieu_aux {xc=a1; yc=b1} {xc=a2; yc=b2} =
     {xc=(a1 +. a2) /. 2.; yc=(b1 +. b2) /. 2. };;

let milieu m1 m2 =
    match (m1,m2) with
    | (Point m'1,Point m'2) -> Point (milieu_aux m'1 m'2)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* mediatrice m1 m2  (ou m1 +? m2)
   retourne la droite mediatrice des deux points m1 et m2 *)
let mediatrice_aux {xc=a1; yc=b1} {xc=a2; yc=b2} =
     {u = a1 -. a2; v = b1 -. b2; 
      h = (-. sqr(a1) +. sqr(a2) -. sqr(b1) +. sqr(b2)) /. 2.};;

let mediatrice m1 m2 =
    match (m1,m2) with
    | (Point m'1,Point m'2) -> Droite (mediatrice_aux m'1 m'2)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;


(* bissectrices d1 d2  (ou d1 **? d2)
   retourne un vecteur [|d'1 ; d'2|] contenant les
   deux droites bissectrices des deux droites d1 et d2 *)
let bissectrices_aux d1 d2 =
    let {u=u1;v=v1;h=h1} = normalise d1
    and {u=u2;v=v2;h=h2} = normalise d2 in
    {u=u1 +. u2; v=v1 +. v2; h=h1 +. h2},
    {u=u1 -. u2; v=v1 -. v2; h=h1 -. h2};;

let bissectrices d1 d2 =
    match (d1,d2) with
    | (Droite d'1,Droite d'2) ->
        let (delta1,delta2) = bissectrices_aux d'1 d'2 in
        [| Droite delta1; Droite delta2|]
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;


(* projection m obj  (ou m *? obj)
   retourne le point projection du point m sur l'objet obj qui
   peut-etre une droite ou un cercle;
   les deux parametres peuvent etre echanges sans inconvenient *)

let projection_p_d {xc=a;yc=b} d =
    let {u=u;v=v;h=h} = normalise d in
    let t = u *. a +. v *. b +. h in
    {xc=a -. t *. u; yc=b -. t *. v};;

let projection_p_c {xc=a0;yc=b0} {xcentre=a;ycentre=b;rayon=r} =
    let norme = sqrt((sqr (a -. a0)) +. (sqr (b -. b0))) in
    {xc=a +. (a0 -. a) *. r /. norme; yc=b +. (b0 -. b) *. r /. norme};;

let projection o1 o2 =
    match (o1,o2) with
    | (Point m, Droite d) -> Point (projection_p_d m d)
    | (Droite d, Point m) -> Point (projection_p_d m d)
    | (Point m, Cercle c) -> Point (projection_p_c m c)
    | (Cercle c, Point m) -> Point (projection_p_c m c)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;




(* symetrique obj1 obj2  (ou  obj1 /? obj2)
   renvoie l'objet  symetrique de obj1 (point, cercle ou droite)
   par rapport a obj2 (point, cercle ou droite);
   la symetrie par rapport a un cercle est l'inversion qui laisse
   stable ce cercle *)
let symetrique_d_d  d1 d2 =
    let {u=u1;v=v1;h=h1} = normalise d1
    and {u=u2;v=v2;h=h2} = normalise d2 in
    let lambda = u1 *. u2 +. v1 *. v2 in
    { u=u1 +. lambda *. u2; v=v1 +. lambda *. v2; h=h1 +. lambda *. h2 };;

let symetrique_p_d {xc=a;yc=b} d =
     let {u=u;v=v;h=h} = normalise d in
     let t = u *. a +. v *. b +. h in
     {xc=a -. 2. *. t *. u; yc=b -. 2. *. t *. v};;

let symetrique_c_d  {xcentre=a;ycentre=b;rayon=r} d =
    cercle_par_centre_et_rayon_aux (symetrique_p_d {xc=a; yc=b} d) r;;

let symetrique_p_p {xc=a1; yc=b1} {xc=a2; yc=b2} =
    {xc=2. *. a2 -. a1;yc=2. *. b2 -. b1};;

let symetrique_d_p  {u=u;v=v;h=h} {xc=a;yc=b} =
    {u=u; v=v; h= -. h +. 2. *. u *. a +. 2. *. v *. b};;

let symetrique_c_p {xcentre=a;ycentre=b;rayon=r} p =
    cercle_par_centre_et_rayon_aux (symetrique_p_p {xc=a; yc=b} p) r;;

let symetrique_p_c {xc=a0; yc=b0} {xcentre=a; ycentre=b; rayon=r} =
    let norme2 = sqr (a -. a0) +. sqr (b -. b0) in
    {xc = a +. r*. (a0 -. a) /. norme2; yc = b +. r*. (b0 -. b) /. norme2};;

let symetrique_d_c {u=u;v=v;h=h} {xcentre=a; ycentre=b; rayon=r}=
    let den = u *. a +. v *. b +. h in
    let a1 = (-. v *. b *. a -. h *. a -. u *. a *. a +. u *. r /. 2.) /. den
    and b1 = (-. h *. b -. v *. b *. b -. u *. a *. b +. v *. r /. 2.) /. den
    and c1 = 
    (v *. b *. b *. b +. u *. a *. b *. b +.
     u *. a *. a *. a -.
     u *. r *. a +. h *. b *. b +. v *. b *. a *. a +.
     h *. a *. a -. v *. r *. b) /. den
    in {xcentre= -. a1; ycentre= -. b1;
        rayon= sqrt(a1 *. a1 +. b1 *. b1 -. c1)};;

let symetrique_c_c
     {xcentre=a1;ycentre=b1;rayon=r1} {xcentre=a2;ycentre=b2;rayon=r2} =
    let den =
      -. a2 *. a2 +. 2.0 *. a2 *. a1 +. 2.0 *. b2 *. b1
      -. a1 *. a1 +. r1 *. r1 -. b2 *. b2 -. b1 *. b1 in
    let a0 =
      -. (-. r2 *. a1 +. r2 *. a2 -. a2 *. b2 *. b2 +. 2.0 *. a1 *. a2 *. a2
          -. a2 *. a2 *. a2 -. b1 *. b1 *. a2
          +. r1 *. r1 *. a2 -. a1 *. a1 *. a2
          +. 2.0 *. b2 *. b1 *. a2) /. den
    and b0 =
      -. ( -. b2 *. b2 *. b2 -. r2 *. b1 -. b1 *. b1 *. b2 +. b2 *. r2
           -. a1 *. a1 *. b2 +. 2.0 *. a2 *. a1 *. b2 +. 2.0 *. b2 *. b2 *. b1
           -. a2 *. a2 *. b2 +. r1 *. r1 *. b2) /. den
    and c0 =
      -. (a2 *. a2 *. a2 *. a2 +. b2 *. b2 *. b2 *. b2 +. r2 *. r2
          +. 2.0 *. a2 *. a2 *. b2 *. b2 -. r1 *. r1 *. b2 *. b2
          -. 2.0 *. a2 *. a1 *. b2 *. b2 -. 2.0 *. b2 *. b2 *. b2 *. b1
          +. b1 *. b1 *. b2 *. b2 -. 2.0 *. b2 *. b2 *. r2
          -. 2.0 *. b2 *. b1 *. a2 *. a2
          +. a1 *. a1 *. b2 *. b2 +. a1 *. a1 *. a2 *. a2
          +. 2.0 *. r2 *. a1 *. a2
          -. r1 *. r1 *. a2 *. a2 +. b1 *. b1 *. a2 *. a2
          -. 2.0 *. a2 *. a2 *. a2 *. a1
          -. 2.0 *. a2 *. a2 *. r2 +. 2.0 *. r2 *. b2 *. b1) /. den
    in {xcentre= -. a0; ycentre= -. b0;
        rayon= sqrt(a0 *. a0 +. b0 *. b0 -. c0)};;

let symetrique ob1 ob2 =
    match (ob1,ob2) with
    | (Point m1,Point m2) -> Point (symetrique_p_p m1 m2)
    | (Droite d1,Point m2) -> Droite (symetrique_d_p d1 m2)
    | (Cercle c1,Point m2) -> Cercle (symetrique_c_p c1 m2)
    | (Point m1,Droite d2) -> Point (symetrique_p_d m1 d2)
    | (Droite d1,Droite d2) -> Droite (symetrique_d_d d1 d2)
    | (Cercle c1,Droite d2) ->  Cercle (symetrique_c_d c1 d2)
    | (Point m1,Cercle c2) -> Point (symetrique_p_c m1 c2)
    | (Droite d1,Cercle c2) -> Cercle (symetrique_d_c d1 c2)
    | (Cercle c1,Cercle c2) ->  Cercle (symetrique_c_c c1 c2)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* centre_de_gravite : objet_geometrique -> objet_geometrique
centre_du_cercle_circonscrit : objet_geometrique -> objet_geometrique
cercle_circonscrit : objet_geometrique -> objet_geometrique
centre_du_cercle_inscrit : objet_geometrique -> objet_geometrique
cercle_inscrit : objet_geometrique -> objet_geometrique
orthocentre : objet_geometrique -> objet_geometrique
cercle_des_9_points : objet_geometrique -> objet_geometrique
droite_Euler : objet_geometrique -> objet_geometrique

ces fonctions recoivent un triangle comme parametre et renvoient l'objet
geometrique correspondant *)

let centre_de_gravite_aux {xc=a1; yc=b1} {xc=a2; yc=b2} {xc=a3; yc=b3} =
    { xc=(a1 +. a2 +. a3) /. 3.; yc=(b1 +. b2 +. b3) /. 3.};;

let centre_de_gravite = function
    | (Triangle (m1,m2,m3))  -> Point (centre_de_gravite_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let centre_du_cercle_circonscrit_aux m1 m2 m3 =
    intersection_d_d (mediatrice_aux m1 m2) (mediatrice_aux m2 m3);;

let centre_du_cercle_circonscrit = function
    | (Triangle (m1,m2,m3)) -> Point (centre_du_cercle_circonscrit_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let cercle_circonscrit = function
    | (Triangle (m1,m2,m3)) ->
         Cercle (cercle_par_centre_et_point_aux
                  (centre_du_cercle_circonscrit_aux m1 m2 m3) m1)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let centre_du_cercle_inscrit_aux
     ({xc=a1; yc=b1} as m1) ({xc=a2; yc=b2} as m2) ({xc=a3; yc=b3} as m3) =
    let d1 = distance m2 m3
    and d2 = distance m3 m1
    and d3 = distance m1 m2 in
    let d = d1 +. d2 +. d3 in
    {xc = (d1 *. a1 +. d2 *. a2 +. d3 *. a3) /. d;
     yc = (d1 *. b1 +. d2 *. b2 +. d3 *. b3) /. d};;

let centre_du_cercle_inscrit  = function
    | (Triangle (m1,m2,m3)) ->
         Point (centre_du_cercle_inscrit_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let cercle_inscrit_aux m1 m2 m3 =
    let omega = centre_du_cercle_inscrit_aux m1 m2 m3
    and m2m3 = droite_par_2_points_aux m2 m3 in
    cercle_par_centre_et_point_aux omega (projection_p_d omega m2m3);;

let cercle_inscrit  = function
    | (Triangle (m1,m2,m3)) ->
         Cercle (cercle_inscrit_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let orthocentre_aux m1 m2 m3 =
    let m2m3 = droite_par_2_points_aux m2 m3 and
    m1m2 = droite_par_2_points_aux m1 m2 in
    intersection_d_d (droite_perpendiculaire_aux m1m2 m3)
                     (droite_perpendiculaire_aux m2m3 m1);;

let orthocentre = function
    | (Triangle (m1,m2,m3)) -> Point (orthocentre_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let cercle_des_9_points_aux m1 m2 m3 =
  let m2m3 = droite_par_2_points_aux m2 m3
  and m3m1 = droite_par_2_points_aux m3 m1
  and m1m2 = droite_par_2_points_aux m1 m2 in
  let h1 = projection_p_d m1 m2m3 and h2 = projection_p_d m2 m3m1
  and h3 = projection_p_d m3 m1m2 in
  cercle_par_centre_et_point_aux
    (centre_du_cercle_circonscrit_aux h1 h2 h3) h1;;

let cercle_des_9_points = function
    | (Triangle (m1,m2,m3)) ->
         Cercle (cercle_des_9_points_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let droite_Euler_aux m1 m2 m3 =
    droite_par_2_points_aux
     (centre_de_gravite_aux m1 m2 m3) (orthocentre_aux m1 m2 m3);;

let droite_Euler  = function
    | (Triangle (m1,m2,m3)) ->
         Droite (droite_Euler_aux m1 m2 m3)
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;



(* ces fonctions recoivent un triangle comme parametre et renvoient un vecteur
   dont les trois composantes sont les objets geometriques correspondants
 *)
let centres_des_cercles_exinscrits_aux
   ({xc=a1; yc=b1} as m1) ({xc=a2; yc=b2} as m2) ({xc=a3; yc=b3} as m3) =
   let d1 = distance m2 m3
   and d2 = distance m3 m1
   and d3 = distance m1 m2 in
   let s3 = d1 +. d2 -. d3
   and s2 = d1 +. d3 -. d2
   and s1 = d2 +. d3 -. d1 in
   ({xc = (d1 *. a1 +. d2 *. a2 -. d3 *. a3) /. s3;
     yc = (d1 *. b1 +. d2 *. b2 -. d3 *. b3) /. s3},
    {xc = (d1 *. a1 -. d2 *. a2 +. d3 *. a3) /. s2;
     yc = (d1 *. b1 -. d2 *. b2 +. d3 *. b3) /. s2},
    {xc = (d2 *. a2 +. d3 *. a3 -. d1 *. a1) /. s1;
     yc = (d2 *. b2 +. d3 *. b3 -. d1 *. b1) /. s1});;

let centres_des_cercles_exinscrits  = function
    | (Triangle (m1,m2,m3)) ->
         let (omega1,omega2,omega3) =
           centres_des_cercles_exinscrits_aux m1 m2 m3 in
         [| Point omega1; Point omega2 ; Point omega3 |]
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let cercles_exinscrits_aux m1 m2 m3 =
   let (omega1,omega2,omega3) = centres_des_cercles_exinscrits_aux m1 m2 m3
   and m2m3 = droite_par_2_points_aux m2 m3
   and m1m2 = droite_par_2_points_aux m1 m2
   and m3m1 = droite_par_2_points_aux m1 m3 in
   (cercle_par_centre_et_point_aux omega1 (projection_p_d omega1 m2m3),
    cercle_par_centre_et_point_aux omega2 (projection_p_d omega2 m3m1),
    cercle_par_centre_et_point_aux omega3 (projection_p_d omega3 m1m2));;

let cercles_exinscrits  = function
    | (Triangle (m1,m2,m3)) ->
         let (c1,c2,c3) = cercles_exinscrits_aux m1 m2 m3 in
         [| Cercle c1; Cercle c2; Cercle c3 |]
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

(* appartient_tol eps m obj
   teste si un point m appartient (approximativement) a
   l'objet obj (une droite ou un cercle) avec la tolerance eps *)
let appartient_tol tolerance m obj =
    match (m,obj) with
    | (Point {xc=x; yc=y},Droite {u=u; v=v; h=h}) ->
       abs_float(u *. x +. v *. y +. h) <. tolerance
    | (Point {xc=x; yc=y},Cercle {xcentre=xc ; ycentre=yc; rayon =r}) ->
       abs_float (sqr(x -. xc) +. sqr(y -. yc) -. sqr(r)) <. tolerance
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;


let appartient = appartient_tol 1e-6;;


(* appartient_tol eps m obj
   teste si un point m appartient (approximativement) a l'objet obj
   (une droite ou un cercle) avec la tolerance eps *)

let alignes_tol tolerance m1 m2 m3 =
    match (m1,m2,m3) with
    | (Point {xc=x1;yc=y1},Point {xc=x2;yc=y2},Point {xc=x3;yc=y3}) ->
        abs_float((y1 -. y2) *. (x2 -. x3)
                  -. (x1 -. x2) *. (y2 -. y3)) <. tolerance
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let alignes = alignes_tol 1e-6;;


(* concourantes_tol eps d1 d2 d3
   teste si les trois droites d1, d2 et d3
   sont (approximativement) concourantes avec la tolerance eps *)
let concourantes_tol tolerance d1 d2 d3 =
    match (d1,d2,d3) with
    | (Droite {u=u1; v=v1; h=h1},Droite {u=u2; v=v2; h=h2 },
       Droite {u=u3; v=v3; h=h3}) ->
         abs_float(h1 *. (v3 *. u2 -. v2 *. u3) +.
                   h2 *. (v1 *. u3 -. v3 *. u1) +.
                   h3 *. (v2 *. u1 -. v1 *. u2)) <. tolerance
    | _ -> raise (Erreur "non defini sur ce type d'objet(s)");;

let concourantes = concourantes_tol 1e-6;;


(* fenetre param x1 x2 y1 y2 ouvre une fenetre graphique en passant
param a la fonction open_graph et definit la fenetre mathematique par
les abscisses x1,x2 et les ordonnees y1,y2. A vous d'ajuster ces
parametres pour que le repere soit orthonorme. *)

let fenetre_gauche = ref 0.0
and fenetre_droite = ref 1.0
and fenetre_bas = ref 0.0
and fenetre_haut = ref 1.0
and rapport_h = ref  0.
and rapport_v = ref  0.;;

let fenetre s gauche droite bas haut =
    open_graph s;
    clear_graph ();
    fenetre_gauche := gauche; fenetre_droite := droite;
    fenetre_haut := haut; fenetre_bas := bas;
    rapport_h := (float_of_int (size_x ())) /. (droite -. gauche);
    rapport_v := (float_of_int (size_y ())) /. (haut -. bas);;

(* fonctions de correspondance math -> ecran *)
let ajuste_h x = int_of_float ((x -. !fenetre_gauche)*. !rapport_h)
and ajuste_v y = int_of_float ((y -. !fenetre_bas)*. !rapport_v)
and ajuste_lh x = int_of_float (x *. !rapport_h)
and ajuste_lv y = int_of_float (y *. !rapport_v);;


(* dessinent un objet ou une liste d'objets *)
let trace_point {xc=a; yc=b} =
    let x = ajuste_h a
    and y =ajuste_v b in
    moveto (x-2) y; lineto (x+2) y;
    moveto x (y-2); lineto x (y+2)

and trace_cercle {xcentre=a; ycentre=b; rayon =r} =
    draw_ellipse (ajuste_h a) (ajuste_v b) (ajuste_lh r) (ajuste_lv r)

and trace_droite {u=u; v=v; h=h} =
    if abs_float v >. abs_float u then
       begin
         moveto 0 (ajuste_v ( -. (u *. !fenetre_gauche +. h) /. v));
         lineto (size_x ()) (ajuste_v ( -. (u *. !fenetre_droite +. h) /. v))
       end
    else
      begin
         moveto (ajuste_h ( -. (v *. !fenetre_bas +. h) /. u)) 0;
         lineto (ajuste_h ( -. (v *. !fenetre_haut +. h) /. u)) (size_y ())
      end

and trace_triangle {xc=a1;yc=b1} {xc=a2;yc=b2} {xc=a3;yc=b3} =
    moveto (ajuste_h a1) (ajuste_v b1);
    lineto (ajuste_h a2) (ajuste_v b2);
    lineto (ajuste_h a3) (ajuste_v b3);
    lineto (ajuste_h a1) (ajuste_v b1);;

let dessine_objet = function
    | Point m -> trace_point m
    | Droite d -> trace_droite d
    | Cercle c -> trace_cercle c
    | Triangle (m1,m2,m3) -> trace_triangle m1 m2 m3
    | Couleur x -> set_color x
    | Epaisseur x -> set_line_width x;;

let rec dessine = function
    | [] -> ()
    | h :: t -> dessine_objet h; dessine t;;


(* dfinition des oprateurs infixes qui servent de raccourcis *)
let prefix *? ob1 ob2 =
    match (ob1,ob2) with
    | (Point m1,Point m2) -> Droite (droite_par_2_points_aux m1 m2)
    | (Point m1,Droite d2) -> Point (projection_p_d m1 d2)
    | (Point m1,Cercle c2) -> Point (projection_p_c m1 c2)
    | (Droite d1,Point m2) -> Droite (droite_parallele_aux d1 m2)
    | (Droite d1,Droite d2) -> Point (intersection_d_d d1 d2)
    | _ -> raise (Erreur " *? non defini sur ces objets");;

let prefix //? ob1 ob2 =
    match (ob1,ob2) with
    | (Point m,Droite d) -> Droite (droite_parallele_aux d m)
    | (Droite d,Point m) -> Droite (droite_parallele_aux d m)
    | _ -> raise (Erreur " //? non defini sur ces objets");;


let prefix **? ob1 ob2 =
    match (ob1,ob2) with
    | (Droite d1, Droite d2) ->
         let (delta1,delta2) = bissectrices_aux d1 d2 in
         [| Droite delta1; Droite delta2|]
    | (Cercle c1, Droite d2) ->
         let (m1,m2) = intersection_d_c d2 c1 in
         [| Point m1; Point m2 |]
    | (Droite d1, Cercle c2) ->
         let (m1,m2) = intersection_d_c d1 c2 in
         [| Point m1; Point m2 |]
    | (Cercle c1, Cercle c2) ->
         let (m1,m2) = intersection_c_c c1 c2 in
         [| Point m1; Point m2 |]
    | _ -> raise (Erreur "**? non defini sur ces objets ");;

let prefix +? ob1 ob2 =
    match (ob1,ob2) with
    | (Point m1,Point m2) -> Droite (mediatrice_aux m1 m2)
    | (Point m1,Droite d2) -> Droite (droite_perpendiculaire_aux d2 m1)
    | (Droite d1,Point m2) -> Droite (droite_perpendiculaire_aux d1 m2)
    | (Cercle c1,Cercle c2) -> Droite (axe_radical_aux c1 c2)
    | _ -> raise (Erreur "+? non defini sur ces objets ");;

let prefix /? ob1 ob2 =
    match (ob1,ob2) with
    | (Point m1,Point m2) -> Point (symetrique_p_p m1 m2)
    | (Droite d1,Point m2) -> Droite (symetrique_d_p d1 m2)
    | (Cercle c1,Point m2) -> Cercle (symetrique_c_p c1 m2)
    | (Point m1,Droite d2) -> Point (symetrique_p_d m1 d2)
    | (Droite d1,Droite d2) -> Droite (symetrique_d_d d1 d2)
    | (Cercle c1,Droite d2) ->  Cercle (symetrique_c_d c1 d2)
    | (Point m1,Cercle c2) -> Point (symetrique_p_c m1 c2)
    | (Droite d1,Cercle c2) -> Cercle (symetrique_d_c d1 c2)
    | (Cercle c1,Cercle c2) ->  Cercle (symetrique_c_c c1 c2)
    | _ -> raise (Erreur " *? non defini sur ces objets");;
